<div id="page-notification">
<i class="fa fa-info-circle" aria-hidden="true"></i>
Sigmund is currently in public beta. You may encounter issues.
</div>
<div id="header-bar">
    <img src='/static/sofa-with-sigmund.png' id="header-logo">
</div>
